import freetype
import logging
from .exceptions import NameNotFoundException
from enum import IntEnum
from io import BufferedReader
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from fontTools.ttLib.ttFont import TTFont
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable
from fontTools.ttLib.tables._n_a_m_e import NameRecord
from fontTools.varLib.instancer.names import ELIDABLE_AXIS_VALUE_NAME
from struct import error as struct_error

_logger = logging.getLogger(__name__)


class NameID(IntEnum):
    COPYRIGHT = 0
    FAMILY_NAME = 1
    SUBFAMILY_NAME = 2
    UNIQUE_ID = 3
    FULL_NAME = 4
    VERSION_STRING = 5
    POSTSCRIPT_NAME = 6
    TRADEMARK = 7
    MANUFACTURER = 8
    DESIGNER = 9
    DESCRIPTION = 10
    VENDOR_URL = 11
    DESIGNER_URL = 12
    LICENSE_DESCRIPTION = 13
    LICENSE_URL = 14
    TYPOGRAPHIC_FAMILY_NAME = 16
    TYPOGRAPHIC_SUBFAMILY_NAME = 17
    MAC_FULL_NAME = 18
    SAMPLE_TEXT = 19
    POSTSCRIPT_CID_FINDFONT_NAME = 20
    WWS_FAMILY_NAME = 21
    WWS_SUBFAMILY_NAME = 22
    LIGHT_BACKGROUND = 23
    DARK_BACKGROUND = 24
    VARIATIONS_PREFIX_NAME = 25


class FontParser:
    DEFAULT_WEIGHT = 400
    DEFAULT_ITALIC = False

    CMAP_ENCODING_MAP = {
        1: {  # Mac
            0: "mac_roman",
        },
        3: {  # Microsoft
            0: "utf_16_be",
            1: "utf_16_be",
            2: "cp932",
            3: "cp936",
            4: "cp950",
            5: "cp949",
            6: "cp1361",
            10: "utf_16_be",
        },
    }

    @staticmethod
    def is_truetype(font: TTFont) -> bool:
        """
        Parameters:
            font (TTFont): An fontTools object
        Returns:
            An boolean that indicate if the font is an Truetype font or not.
            If not, then, it is an Opentype font.
        """
        return "glyf" in font

    @staticmethod
    def is_valid_variable_font(font: TTFont) -> bool:
        """
        Parameters:
            font (TTFont): An fontTools object
        Returns:
            An boolean that indicate if the font is an variable font or not.
        """

        if "fvar" not in font or "STAT" not in font:
            return False

        if font["STAT"].table is None:
            return False

        for axe in font["fvar"].axes:
            if not (axe.minValue <= axe.defaultValue <= axe.maxValue):
                return False

        if font["STAT"].table.AxisValueArray is not None:
            for axis_value in font["STAT"].table.AxisValueArray.AxisValue:
                if axis_value.Format in (1, 2, 3, 4):
                    if axis_value.Format == 4 and len(axis_value.AxisValueRecord) == 0:
                        return False
                else:
                    return False

        return True

    @staticmethod
    def get_var_font_family_prefix(font: TTFont) -> str:
        """
        From: https://github.com/fonttools/fonttools/blob/3c4cc71504774d1ae4f1e59e3ef3b97e194c1c91/Lib/fontTools/varLib/instancer/names.py#L267-L269

        Parameters:
            font (TTFont): An fontTools object
        Returns:
            The family name prefix.
        """
        try:
            family_prefix = FontParser.get_name_by_id(
                NameID.TYPOGRAPHIC_FAMILY_NAME, font["name"].names
            )
        except NameNotFoundException:
            family_prefix = FontParser.get_name_by_id(
                NameID.FAMILY_NAME, font["name"].names
            )

        return family_prefix

    @staticmethod
    def get_distance_between_axis_value_and_coordinates(
        font: TTFont, coordinates: Dict[str, float], axis_value: Any, axis_format: int
    ) -> float:
        """
        Parameters:
            ttfont (TTFont): An fontTools object
            coordinates (Dict[str, float]): The coordinates of an NamedInstance in the fvar table.
            axis_value (Any): An AxisValue
            axis_format (int): The AxisValue Format.
                Since the AxisValue from AxisValueRecord of an AxisValue Format 4 doesn't contain an Format attribute, this parameter is needed.
        Returns:
            The distance between_axis_value_and_coordinates
        """

        axis_tag = (
            font["STAT"].table.DesignAxisRecord.Axis[axis_value.AxisIndex].AxisTag
        )
        instance_value = coordinates.get(axis_tag, 0)

        if axis_format == 2:
            clamped_axis_value = max(
                min(instance_value, axis_value.RangeMaxValue),
                axis_value.RangeMinValue,
            )
        else:
            clamped_axis_value = axis_value.Value

        delta = clamped_axis_value - instance_value
        delta_square = delta**2

        if delta < 0:
            adjust = 1
        else:
            adjust = 0

        distance = delta_square * 2 + adjust

        return distance

    @staticmethod
    def get_axis_value_from_coordinates(
        ttfont: TTFont, coordinates: Dict[str, float]
    ) -> List[Any]:
        """
        Parameters:
            ttfont (TTFont): An fontTools object
            coordinates (Dict[str, float]): The coordinates of an NamedInstance in the fvar table.
        Returns:
            An list who contain all the AxisValue linked to the coordinates.
        """

        distances_for_axis_values: List[Tuple[float, Any]] = []

        if ttfont["STAT"].table.AxisValueArray is None:
            return distances_for_axis_values

        for axis_value in ttfont["STAT"].table.AxisValueArray.AxisValue:

            if axis_value.Format == 4:
                distance = 0

                for axis_value_format_4 in axis_value.AxisValueRecord:
                    distance += (
                        FontParser.get_distance_between_axis_value_and_coordinates(
                            ttfont, coordinates, axis_value_format_4, axis_value.Format
                        )
                    )

                distances_for_axis_values.append((distance, axis_value))

            else:
                distance = FontParser.get_distance_between_axis_value_and_coordinates(
                    ttfont, coordinates, axis_value, axis_value.Format
                )
                distances_for_axis_values.append((distance, axis_value))

        # Sort by ASC
        distances_for_axis_values.sort(key=lambda distance: distance[0])

        axis_values_coordinate_matches: List[Any] = []
        is_axis_useds: List[bool] = [False] * len(
            ttfont["STAT"].table.DesignAxisRecord.Axis
        )

        for distance, axis_value in distances_for_axis_values:
            if axis_value.Format == 4:

                # The AxisValueRecord can have "internal" duplicate axis, but it cannot have duplicate Axis with the other AxisValue
                is_any_duplicate_axis = False
                for axis_value_format_4 in axis_value.AxisValueRecord:
                    if is_axis_useds[axis_value_format_4.AxisIndex]:
                        is_any_duplicate_axis = True
                        break

                if not is_any_duplicate_axis:
                    for axis_value_format_4 in axis_value.AxisValueRecord:
                        is_axis_useds[axis_value_format_4.AxisIndex] = True

                    axis_values_coordinate_matches.append(axis_value)
            else:
                if not is_axis_useds[axis_value.AxisIndex]:
                    is_axis_useds[axis_value.AxisIndex] = True
                    axis_values_coordinate_matches.append(axis_value)

        return axis_values_coordinate_matches

    @staticmethod
    def get_axis_value_table_property(
        ttfont: TTFont, axis_values: List[Any], family_name_prefix: str
    ) -> Tuple[str, str, float, bool]:
        """
        Parameters:
            ttfont (TTFont): An fontTools object
            axis_values (List[Any]): An list of AxisValue.
            family_name_prefix (str): The variable family name prefix.
                Ex: For the name "Alegreya Italic", "Alegreya" is the family name prefix.
        Returns:
            An FontFace that represent the axis_values.
        """

        axis_values.sort(
            key=lambda axis_value: ttfont["STAT"]
            .table.DesignAxisRecord.Axis[
                min(
                    axis_value.AxisValueRecord,
                    key=lambda axis_value_format_4: ttfont["STAT"]
                    .table.DesignAxisRecord.Axis[axis_value_format_4.AxisIndex]
                    .AxisOrdering,
                ).AxisIndex
            ]
            .AxisOrdering
            if axis_value.Format == 4
            else ttfont["STAT"]
            .table.DesignAxisRecord.Axis[axis_value.AxisIndex]
            .AxisOrdering
        )

        family_axis_value = []
        fullname_axis_value = []
        weight = FontParser.DEFAULT_WEIGHT
        italic = FontParser.DEFAULT_ITALIC

        for axis_value in axis_values:

            # If the Format 4 only contain only 1 AxisValueRecord, it will treat it as an single AxisValue like the Format 1, 2 or 3.
            if axis_value.Format == 4 and len(axis_value.AxisValueRecord) > 1:
                if not axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME:
                    family_axis_value.append(axis_value)
                    fullname_axis_value.append(axis_value)
            else:
                if axis_value.Format == 2:
                    value = axis_value.NominalValue
                    axis_index = axis_value.AxisIndex
                elif axis_value.Format in (1, 3):
                    value = axis_value.Value
                    axis_index = axis_value.AxisIndex
                elif axis_value.Format == 4:
                    value = axis_value.AxisValueRecord[0].Value
                    axis_index = axis_value.AxisValueRecord[0].AxisIndex

                if (
                    ttfont["STAT"].table.DesignAxisRecord.Axis[axis_index].AxisTag
                    == "wght"
                ):
                    weight = value
                elif (
                    ttfont["STAT"].table.DesignAxisRecord.Axis[axis_index].AxisTag
                    == "ital"
                ):
                    italic = value == 1

                if not (axis_value.Flags & ELIDABLE_AXIS_VALUE_NAME):
                    fullname_axis_value.append(axis_value)

                    use_in_family_name = True
                    if (
                        ttfont["STAT"].table.DesignAxisRecord.Axis[axis_index].AxisTag
                        == "wght"
                    ):
                        use_in_family_name = value not in (400, 700)
                    elif (
                        ttfont["STAT"].table.DesignAxisRecord.Axis[axis_index].AxisTag
                        == "ital"
                    ):
                        use_in_family_name = value not in (0, 1)

                    if use_in_family_name:
                        family_axis_value.append(axis_value)

        try:
            family_name = f'{family_name_prefix} {" ".join(FontParser.get_name_by_id(axis_value.ValueNameID, ttfont["name"].names) for axis_value in family_axis_value)}'.strip(
                " "
            )
        except NameNotFoundException:
            family_name = family_name_prefix

        if len(fullname_axis_value) == 0:
            try:
                full_name = f"{family_name_prefix} {FontParser.get_name_by_id(ttfont['STAT'].table.ElidedFallbackNameID, ttfont['name'].names)}".strip(
                    " "
                )
            except NameNotFoundException:
                weight = FontParser.DEFAULT_WEIGHT
                italic = FontParser.DEFAULT_ITALIC
                full_name = f"{family_name_prefix} Regular"

        else:
            try:
                full_name = f'{family_name_prefix} {" ".join(FontParser.get_name_by_id(axis_value.ValueNameID, ttfont["name"].names) for axis_value in fullname_axis_value)}'.strip(
                    " "
                )
            except NameNotFoundException:
                weight = FontParser.DEFAULT_WEIGHT
                italic = FontParser.DEFAULT_ITALIC

                try:
                    full_name = f"{family_name_prefix} {FontParser.get_name_by_id(ttfont['STAT'].table.ElidedFallbackNameID, ttfont['name'].names)}".strip(
                        " "
                    )
                except NameNotFoundException:
                    full_name = f"{family_name_prefix} Regular"

        return family_name, full_name, weight, italic

    @staticmethod
    def sort_naming_table(names: List[NameRecord]) -> List[NameRecord]:
        """
        Parameters:
            names (List[NameRecord]): Naming table
        Returns:
            The sorted naming table
        """

        def is_english(name: NameRecord) -> bool:
            # From https://gitlab.freedesktop.org/fontconfig/fontconfig/-/blob/d863f6778915f7dd224c98c814247ec292904e30/src/fcfreetype.c#L1111-1125
            return (name.platformID, name.langID) in ((1, 0), (3, 0x409))

        # From	https://gitlab.freedesktop.org/fontconfig/fontconfig/-/blob/d863f6778915f7dd224c98c814247ec292904e30/src/fcfreetype.c#L1078
        # 		https://github.com/freetype/freetype/blob/b98dd169a1823485e35b3007ce707a6712dcd525/include/freetype/ttnameid.h#L86-L91
        PLATFORM_ID_APPLE_UNICODE = 0
        PLATFORM_ID_MACINTOSH = 1
        PLATFORM_ID_ISO = 2
        PLATFORM_ID_MICROSOFT = 3
        PLATFORM_ID_ORDER = [
            PLATFORM_ID_MICROSOFT,
            PLATFORM_ID_APPLE_UNICODE,
            PLATFORM_ID_MACINTOSH,
            PLATFORM_ID_ISO,
        ]

        return sorted(
            names,
            key=lambda name: (
                PLATFORM_ID_ORDER.index(name.platformID),
                name.nameID,
                name.platEncID,
                -is_english(name),
                name.langID,
            ),
        )

    @staticmethod
    def get_name_by_id(nameID: int, names: List[NameRecord]) -> str:
        """
        Parameters:
            nameID (int): ID of the name you search
            names (List[NameRecord]): Naming table
        Returns:
            The decoded name
        """

        names = list(filter(lambda name: name.nameID == nameID, names))
        names = FontParser.sort_naming_table(names)

        for name in names:
            try:
                name_str = FontParser.get_decoded_name(name)
            except UnicodeDecodeError:
                continue

            return name_str

        raise NameNotFoundException(
            f"The NamingTable doesn't contain the NameID {nameID}"
        )

    @staticmethod
    def get_decoded_name(name: NameRecord) -> str:
        """
        Parameters:
            names (NameRecord): Name record from the naming record
        Returns:
            The decoded name
        """

        encoding = FontParser.get_name_encoding(name)

        try:
            return name.string.decode(encoding)
        except UnicodeDecodeError:
            utf_16_decoded = name.string.decode("utf_16_be")
            to_decode = bytes(utf_16_decoded, encoding="raw_unicode_escape")
            return to_decode.decode(encoding)

    @staticmethod
    def get_font_postscript_property(font_path: str, font_index: int) -> Optional[str]:
        """
        Parameters:
            font_path (str): Font path.
            font_index (int): Font index.
        Returns:
            The postscript name
        """
        try:
            # Like libass, we use freetype: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L326
            postscriptNameByte = freetype.Face(
                Path(font_path).open("rb"), font_index
            ).postscript_name
        except OSError:
            _logger.warning(
                f'Error: Please report this error on github. Attach this font "{font_path}" in your issue and say that the postscript has not been correctly decoded'
            )

        if postscriptNameByte is not None:
            try:
                postscriptName = postscriptNameByte.decode("ASCII")
            except UnicodeDecodeError:
                _logger.warning(
                    f'Error: Please report this error on github. Attach this font "{font_path}" in your issue and say that the postscript has not been correctly decoded'
                )

            if postscriptName is not None:
                return postscriptName

        return None

    @staticmethod
    def get_font_italic_bold_property(
        font: TTFont, font_path: str, font_index: int
    ) -> Tuple[bool, int]:
        """
        Parameters:
            font (TTFont): An fontTools object
            font_path (str): Font path.
            font_index (int): Font index.
        Returns:
            is_italic, weight
        """

        def get_font_italic_bold_property_with_freetype(
            font_path: str, font_index: int
        ) -> Tuple[bool, int]:
            font = freetype.Face(Path(font_path).open("rb"), font_index)
            # From: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L318
            is_italic = bool(
                font.style_flags & freetype.ft_enums.ft_style_flags.FT_STYLE_FLAG_ITALIC
            )
            # From: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_font.c#L523
            weight = (
                700
                if bool(
                    font.style_flags
                    & freetype.ft_enums.ft_style_flags.FT_STYLE_FLAG_BOLD
                )
                else 400
            )

            return is_italic, weight

        if "OS/2" in font:
            try:
                # https://docs.microsoft.com/en-us/typography/opentype/spec/os2#fss
                is_italic = bool(font["OS/2"].fsSelection & 1)

                # https://docs.microsoft.com/en-us/typography/opentype/spec/os2#usweightclass
                weight = font["OS/2"].usWeightClass
            except struct_error:
                is_italic, weight = get_font_italic_bold_property_with_freetype(
                    font_path, font_index
                )
        else:
            is_italic, weight = get_font_italic_bold_property_with_freetype(
                font_path, font_index
            )

        # See https://github.com/libass/libass/pull/676
        if weight == 1:
            weight = 100
        elif weight == 2:
            weight = 200
        elif weight == 3:
            weight = 300
        elif weight == 4:
            weight = 350
        elif weight == 5:
            weight = 400
        elif weight == 6:
            weight = 600
        elif weight == 7:
            weight = 700
        elif weight == 8:
            weight = 800
        elif weight == 9:
            weight = 900

        return is_italic, weight

    @staticmethod
    def get_font_family_fullname_property(
        names: List[NameRecord],
    ) -> Tuple[Set[str], Set[str]]:
        """
        Parameters:
            names (List[NameRecord]): Naming table
        Returns:
            All decoded families and fullnames that are from microsoft platform
        """

        # https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L258-L344
        MAX = 100
        families = set()
        fullnames = set()

        for name in names:

            if name.platformID == 3 and (
                name.nameID == NameID.FAMILY_NAME or name.nameID == NameID.FULL_NAME
            ):

                try:
                    name_str = FontParser.get_decoded_name(name)
                except UnicodeDecodeError:
                    continue

                if name.nameID == 1 and len(families) < MAX:
                    families.add(name_str)
                elif name.nameID == 4 and len(fullnames) < MAX:
                    fullnames.add(name_str)

        return families, fullnames

    @staticmethod
    def get_cmap_encoding(cmap_table: CmapSubtable) -> Optional[str]:
        """
        Parameters:
            cmap_table (CmapSubtable): CMAP table
        Returns:
            The cmap codepoint encoding.
            If GDI does not support the platform_id and/or platform_encoding_id, return None.
        """
        return FontParser.CMAP_ENCODING_MAP.get(cmap_table.platformID, {}).get(
            cmap_table.platEncID, None
        )

    @staticmethod
    def get_name_encoding(name: NameRecord) -> Optional[str]:
        """
        Parameters:
            names (NameRecord): Name record from the naming record
        Returns:
            The cmap codepoint encoding.
            If GDI does not support the name, return None.
        """
        # From: https://github.com/MicrosoftDocs/typography-issues/issues/956#issuecomment-1205678068
        if name.platformID == 3:
            if name.platEncID == 3:
                return "cp936"
            elif name.platEncID == 4:
                if name.nameID == 2:
                    return "utf_16_be"
                else:
                    return "cp950"
            elif name.platEncID == 5:
                if name.nameID == 2:
                    return "utf_16_be"
                else:
                    return "cp949"
            else:
                return "utf_16_be"
        elif name.platformID == 1 and name.platEncID == 0:
            # From: https://github.com/libass/libass/issues/679#issuecomment-1442262479
            return "iso-8859-1"

        return None

    @staticmethod
    def is_file_truetype_collection(file: BufferedReader) -> bool:
        file.seek(0)
        return b"ttcf" == file.read(4)

    @staticmethod
    def is_file_truetype(file: BufferedReader) -> bool:
        file.seek(0)
        return b"\x00\x01\x00\x00" == file.read(4)

    @staticmethod
    def is_file_opentype(file: BufferedReader) -> bool:
        file.seek(0)
        return b"OTTO" == file.read(4)

    @staticmethod
    def is_file_font(filepath: str) -> bool:
        with open(filepath, "rb") as fontFile:
            return (
                FontParser.is_file_truetype(fontFile)
                or FontParser.is_file_opentype(fontFile)
                or FontParser.is_file_truetype_collection(fontFile)
            )



import os
from font_collector import AssDocument, Helpers, FontLoader

# Get ass path used for tests
dir_path = os.path.dirname(os.path.realpath(__file__))
path_ass = os.path.join(dir_path, "ass", "Bold italic test.ass")

subtitle = AssDocument.from_file(path_ass)


def test_get_used_font_by_style():
    styles = subtitle.get_used_style()
    styles = list(styles.keys())

    assert len(styles) == 1

    style = styles[0]
    font_collection = FontLoader(
        [os.path.join(dir_path, "fonts", "Raleway")], False
    ).fonts

    font_result = Helpers.get_used_font_by_style(font_collection, style)

    # VSFilter prefer to match to the weight then the italic.
    # If it would have prefer to match the italic, the weight would be 700 and italic would be true.
    assert font_result.font.weight == 900
    assert font_result.font.italic == False



import os
from font_collector import AssDocument, AssStyle, UsageData

# Get ass path used for tests
dir_path = os.path.dirname(os.path.realpath(__file__))
path_ass = os.path.join(dir_path, "ass", "Untitled.ass")

subtitle = AssDocument.from_file(path_ass)


def test_get_style_used():
    styles = subtitle.get_used_style()

    expected_results = {
        AssStyle("Sunny Spells", 700, True): UsageData(set("example"), {1}),
        AssStyle("Akashi", 700, True): UsageData(set("Multiple \\t"), {2}),
        AssStyle("Bodo Amat", 400, True): UsageData(set("Italic"), {3}),
        AssStyle("Bodo Amat", 400, False): UsageData(set("Regular"), {3}),
        AssStyle("Avenir LT 65 Medium", 400, False): UsageData(
            set("Don't match with style"), {4}
        ),
        AssStyle("Asap", 400, True): UsageData(set("\\r override"), {5}),
    }

    assert styles == expected_results



from font_collector.font_parser import FontParser
from fontTools.ttLib.tables._n_a_m_e import NameRecord


def test_fallback_encoding():
    # The string are from the font in this pack: https://github.com/libass/libass/issues/643#issuecomment-1476459274

    # 微软简标宋 - PlatEncID 3.TTF
    name_record = NameRecord()
    name_record.nameID = 1
    name_record.string = b"\x00\xce\x00\xa2\x00\xc8\x00\xed\x00\xbc\x00\xf2\x00\xb1\x00\xea\x00\xcb\x00\xce"
    name_record.platformID = 3
    name_record.platEncID = 3
    name_record.langID = 0
    FontParser.get_decoded_name(name_record) == "微软简标宋"

    # 文鼎中特廣告體 - PlatEncID 4.ttf
    name_record = NameRecord()
    name_record.nameID = 1
    name_record.string = b"\x00\xa4\x00\xe5\x00\xb9\x00\xa9\x00\xa4\x00\xa4\x00\xaf\x00S\x00\xbc\x00s\x00\xa7\x00i\x00\xc5\x00\xe9"
    name_record.platformID = 3
    name_record.platEncID = 4
    name_record.langID = 0
    FontParser.get_decoded_name(name_record) == "文鼎中特廣告體"



import os
import string
from font_collector import Font

dir_path = os.path.dirname(os.path.realpath(__file__))
font_without_os2_table = os.path.join(dir_path, "fonts", "font_mac.TTF")
font_without_axis_value = os.path.join(dir_path, "fonts", "font_without axis_value.ttf")


def test_font_without_os2_table():

    font = Font.from_font_path(font_without_os2_table)
    assert len(font) == 1
    font = font[0]

    assert font.family_names == set(["brushstroke plain"])
    assert font.weight == 400
    assert font.italic == False
    assert font.exact_names == set()
    assert font.is_var == False


def test_font_with_fvar_table_but_without_stat_table():

    font_without_stat_table = os.path.join(
        dir_path, "fonts", "Cabin VF Beta Regular.ttf"
    )

    font = Font.from_font_path(font_without_stat_table)
    assert len(font) == 1
    font = font[0]

    assert font.family_names == set(["cabin vf beta"])
    assert font.weight == 400
    assert font.italic == False
    assert font.exact_names == set(["cabin vf beta regular"])
    assert font.is_var == False


def test_font_without_axis_value():
    fonts = Font.from_font_path(font_without_axis_value)
    expected_fonts = [
        Font(
            font_without_axis_value,
            0,
            ["inter"],
            400,
            False,
            ["inter regular"],
            {"wght": 100.0, "slnt": 0.0},
        )
    ] * 18
    assert fonts == expected_fonts


def test_font_get_missing_glyphs_cmap_encoding_2():

    font_cmap_encoding_2 = os.path.join(dir_path, "fonts", "font_cmap_encoding_2.TTF")

    font = Font.from_font_path(font_cmap_encoding_2)
    assert len(font) == 1
    font = font[0]

    missing_glyphs = font.get_missing_glyphs(
        string.ascii_letters + string.digits + "éｦ&*"
    )
    assert missing_glyphs == set(["é"])


def test_font_get_missing_glyphs_cmap_encoding_mac_platform():

    font = Font.from_font_path(font_without_os2_table)
    assert len(font) == 1
    font = font[0]

    missing_glyphs = font.get_missing_glyphs(
        string.ascii_letters + string.digits + "@é¸"
    )
    assert missing_glyphs == set(["@", "¸"])


def test_variable_font_with_invalid_fvar_axes():

    font_path = os.path.join(dir_path, "variable font tests", "Test #1", "Test #1.ttf")

    font = Font.from_font_path(font_path)
    assert len(font) == 1
    font = font[0]

    expected_font = Font(font_path, 0, ["Advent Pro"], 400, True, ["Advent Pro Italic"])

    assert font == expected_font


def test_variable_font_without_fvar_table():

    font_path = os.path.join(dir_path, "variable font tests", "Test #3", "Test #3.ttf")

    font = Font.from_font_path(font_path)
    assert len(font) == 1
    font = font[0]

    expected_font = Font(font_path, 0, ["Advent Pro"], 300, True, ["Advent Pro Italic"])

    assert font == expected_font


def test_variable_font_with_axis_value_format_4_one_axis_value_record():
    # The AxisValue Format 4 contain only 1 AxisValue in the AxisValueRecord

    font_path = os.path.join(dir_path, "variable font tests", "Test #4", "Test #4.ttf")

    fonts = Font.from_font_path(font_path)
    expected_fonts = [
        Font(
            font_path, 0, ["Alegreya"], 400, True, ["Alegreya Italic"], {"wght": 400.0}
        ),
        Font(
            font_path,
            0,
            ["Alegreya Medium"],
            500,
            True,
            ["Alegreya Medium Italic"],
            {"wght": 500.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya"],
            700,
            True,
            ["Alegreya Bold Italic"],
            {"wght": 700.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Roman numerals"],
            800,
            True,
            ["Alegreya Roman numerals Italic"],
            {"wght": 800.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Black"],
            900,
            True,
            ["Alegreya Black Italic"],
            {"wght": 900.0},
        ),
    ]

    assert fonts == expected_fonts


def test_variable_font_with_axis_value_format_4_multiple_axis_value_record():
    # The AxisValue Format 4 contain multiple AxisValue in the AxisValueRecord

    font_path = os.path.join(dir_path, "variable font tests", "Test #5", "Test #5.ttf")

    fonts = Font.from_font_path(font_path)
    expected_fonts = [
        Font(
            font_path, 0, ["Alegreya"], 400, True, ["Alegreya Italic"], {"wght": 400.0}
        ),
        Font(
            font_path,
            0,
            ["Alegreya Medium"],
            500,
            True,
            ["Alegreya Medium Italic"],
            {"wght": 500.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya"],
            700,
            True,
            ["Alegreya Bold Italic"],
            {"wght": 700.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Roman numerals"],
            400,
            False,
            ["Alegreya Roman numerals"],
            {"wght": 400.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Black"],
            900,
            True,
            ["Alegreya Black Italic"],
            {"wght": 900.0},
        ),
    ]

    assert fonts == expected_fonts


def test_variable_font_with_invalid_elided_fallback_nameid():

    font_path = os.path.join(dir_path, "variable font tests", "Test #6", "Test #6.ttf")

    fonts = Font.from_font_path(font_path)
    expected_fonts = [
        Font(
            font_path,
            0,
            ["Advent Pro"],
            400,
            False,
            ["Advent Pro Regular"],
            {"wght": 400.0},
        )
    ] * 9

    assert fonts == expected_fonts


def test_variable_font_match_with_axis_value_format_4_and_axis_format_1():
    # The NamedInstance "wght" == 800 will take the AxisValue Format 4 weight and the AxisValue Format 1 "ital".

    font_path = os.path.join(dir_path, "variable font tests", "Test #7", "Test #7.ttf")

    fonts = Font.from_font_path(font_path)
    expected_fonts = [
        Font(
            font_path, 0, ["Alegreya"], 400, True, ["Alegreya Italic"], {"wght": 400.0}
        ),
        Font(
            font_path,
            0,
            ["Alegreya Medium"],
            500,
            True,
            ["Alegreya Medium Italic"],
            {"wght": 500.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya"],
            700,
            True,
            ["Alegreya Bold Italic"],
            {"wght": 700.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya ExtraBold"],
            755,
            True,
            ["Alegreya ExtraBold Italic"],
            {"wght": 800.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Black"],
            900,
            True,
            ["Alegreya Black Italic"],
            {"wght": 900.0},
        ),
    ]

    assert fonts == expected_fonts


def test_variable_font_duplicate_font_face():

    font_path = os.path.join(dir_path, "variable font tests", "Test #8", "Test #8.ttf")

    fonts = Font.from_font_path(font_path)
    expected_fonts = [
        Font(
            font_path, 0, ["Alegreya"], 400, True, ["Alegreya Italic"], {"wght": 400.0}
        ),
        Font(
            font_path,
            0,
            ["Alegreya Medium"],
            500,
            True,
            ["Alegreya Medium Italic"],
            {"wght": 500.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya"],
            700,
            True,
            ["Alegreya Bold Italic"],
            {"wght": 700.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Black"],
            900,
            True,
            ["Alegreya Black Italic"],
            {"wght": 900.0},
        ),
        Font(
            font_path,
            0,
            ["Alegreya Black"],
            900,
            True,
            ["Alegreya Black Italic"],
            {"wght": 900.0},
        ),
    ]

    assert fonts == expected_fonts



import logging
import os
from .exceptions import InvalidFontException
from .font_parser import FontParser, NameID
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable
from fontTools.ttLib.ttFont import TTFont
from fontTools.ttLib.ttCollection import TTCollection
from typing import Any, Dict, List, Sequence, Set, Tuple

_logger = logging.getLogger(__name__)


class Font:
    filename: str
    font_index: int
    __family_names: Set[str]
    weight: int
    italic: bool
    __exact_names: Set[
        str
    ]  # if the font is a TrueType, it will be the "full_name". if the font is a OpenType, it will be the "postscript name"
    named_instance_coordinates: Dict[str, float] = {}

    def __init__(
        self,
        filename: str,
        font_index: int,
        family_names: Sequence[str],
        weight: int,
        italic: bool,
        exact_names: Sequence[str],
        named_instance_coordinates: Dict[str, float] = {},
    ):
        self.filename = filename
        self.font_index = font_index
        self.family_names = family_names
        self.weight = weight
        self.italic = italic
        self.exact_names = exact_names
        self.named_instance_coordinates = named_instance_coordinates

    @classmethod
    def from_font_path(cls, font_path: str) -> List["Font"]:
        """
        Parameters:
            font_path (str): Font path. The font can be a .ttf, .otf or .ttc
        Returns:
            An Font object that represent the file at the font_path
        """
        ttFonts: List[TTFont] = []
        fonts: List[Font] = []

        with open(font_path, "rb") as font_file:
            if FontParser.is_file_truetype_collection(font_file):
                ttFonts.extend(TTCollection(font_path).fonts)
            elif FontParser.is_file_truetype(font_file) or FontParser.is_file_opentype(
                font_file
            ):
                ttFonts.append(TTFont(font_path))
            else:
                raise FileExistsError(
                    f'The file "{font_path}" is not a valid font file'
                )

        try:

            # Read font attributes
            for font_index, ttFont in enumerate(ttFonts):

                # If is Variable Font, else "normal" font
                if FontParser.is_valid_variable_font(ttFont):
                    fonts.extend(
                        Font._open_variable_font(ttFont, font_path, font_index)
                    )
                else:
                    try:
                        font = Font._open_normal_font(ttFont, font_path, font_index)
                    except InvalidFontException as e:
                        _logger.info(f"{e}. The font {font_path} will be ignored.")
                        continue

                    fonts.append(font)

        except Exception:
            _logger.error(
                f'An unknown error occurred while reading the font "{font_path}"{os.linesep}Please open an issue on github, share the font and the following error message:'
            )
            raise
        return fonts

    @classmethod
    def _open_normal_font(
        cls, ttFont: TTFont, font_path: str, font_index: int
    ) -> "Font":
        """
        Parameters:
            font (TTFont): An fontTools object
            font_path (str): Font path.
            font_index (int): Font index.
        Returns:
            An Font instance that represent the ttFont
        """

        is_truetype = FontParser.is_truetype(ttFont)

        families, fullnames = FontParser.get_font_family_fullname_property(
            ttFont["name"].names
        )

        # This is something like: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L303-L311
        if len(families) == 0:
            family_name = FontParser.get_name_by_id(
                NameID.FAMILY_NAME, ttFont["name"].names
            )

            if family_name:
                families.add(family_name)
            else:
                raise InvalidFontException(
                    "The font does not contain an valid family name"
                )

        if is_truetype:
            exact_names = fullnames
        else:
            exact_names = set()
            postscript_name = FontParser.get_font_postscript_property(
                font_path, font_index
            )
            if postscript_name is not None:
                exact_names.add(postscript_name)

        is_italic, weight = FontParser.get_font_italic_bold_property(
            ttFont, font_path, font_index
        )

        return cls(
            font_path,
            font_index,
            families,
            weight,
            is_italic,
            exact_names,
        )

    @classmethod
    def _open_variable_font(
        cls, ttFont: TTFont, font_path: str, font_index: int
    ) -> List["Font"]:
        """
        Parameters:
            font (TTFont): An fontTools object
            font_path (str): Font path.
            font_index (int): Font index.
        Returns:
            An list of Font instance that represent the ttFont.
        """
        fonts: List[Font] = []

        family_name_prefix = FontParser.get_var_font_family_prefix(ttFont)
        axis_values_coordinates: List[Tuple[Any, Dict[str, float]]] = []

        for instance in ttFont["fvar"].instances:
            families = set()
            fullnames = set()

            axis_value_table = FontParser.get_axis_value_from_coordinates(
                ttFont, instance.coordinates
            )

            instance_coordinates = instance.coordinates
            for axis_value_coordinates in axis_values_coordinates:
                if axis_value_coordinates[0] == axis_value_table:
                    instance_coordinates = axis_value_coordinates[1]
                    break
            axis_values_coordinates.append((axis_value_table, instance.coordinates))

            (
                family_name,
                full_name,
                weight,
                is_italic,
            ) = FontParser.get_axis_value_table_property(
                ttFont, axis_value_table, family_name_prefix
            )

            families.add(family_name)
            fullnames.add(full_name)

            font = cls(
                font_path,
                font_index,
                families,
                int(weight),
                is_italic,
                fullnames,
                instance_coordinates,
            )
            fonts.append(font)

        return fonts

    @property
    def family_names(self):
        return self.__family_names

    @family_names.setter
    def family_names(self, family_names: List[str]):
        self.__family_names = set([family_name.lower() for family_name in family_names])

    @property
    def exact_names(self):
        return self.__exact_names

    @exact_names.setter
    def exact_names(self, exact_names: List[str]):
        self.__exact_names = set([exact_name.lower() for exact_name in exact_names])

    @property
    def is_var(self):
        return len(self.named_instance_coordinates) > 0

    def __eq__(self, other: "Font"):
        return (self.family_names, self.weight, self.italic, self.exact_names) == (
            other.family_names,
            other.weight,
            other.italic,
            other.exact_names,
        )

    def __hash__(self):
        return hash(
            (
                self.filename,
                self.font_index,
                tuple(self.family_names),
                self.italic,
                self.weight,
                tuple(self.exact_names),
            )
        )

    def __repr__(self):
        return f'Filename: "{self.filename}" Family_names: "{self.family_names}", Weight: "{self.weight}", Italic: "{self.italic}, Exact_names: "{self.exact_names}", Named_instance_coordinates: "{self.named_instance_coordinates}"'

    def get_missing_glyphs(self, text: Sequence[str]) -> Set[str]:
        """
        Parameters:
            text (Sequence[str]): Text
        Returns:
            A set of all the character that the font cannot display.
        """

        ttFont = TTFont(self.filename, fontNumber=self.font_index)
        char_not_found: Set[str] = set()

        cmap_tables: List[CmapSubtable] = list(
            filter(lambda table: table.platformID == 3, ttFont["cmap"].tables)
        )

        # GDI seems to take apple cmap if there isn't any microsoft cmap: https://github.com/libass/libass/issues/679
        if len(cmap_tables) == 0:
            cmap_tables = list(
                filter(
                    lambda table: table.platformID == 1 and table.platEncID == 0,
                    ttFont["cmap"].tables,
                )
            )

        for char in text:
            char_found = False

            for cmap_table in cmap_tables:
                cmap_encoding = FontParser.get_cmap_encoding(cmap_table)

                # Cmap isn't supported
                if cmap_encoding is None:
                    continue

                try:
                    codepoint = int.from_bytes(char.encode(cmap_encoding), "big")
                except UnicodeEncodeError:
                    continue

                # GDI/Libass modify the codepoint for microsoft symbol cmap: https://github.com/libass/libass/blob/04a208d5d200360d2ac75f8f6cfc43dd58dd9225/libass/ass_font.c#L249-L250
                if cmap_table.platformID == 3 and cmap_table.platEncID == 0:
                    codepoint = 0xF000 | codepoint

                if codepoint in cmap_table.cmap:
                    char_found = True
                    break

            if not char_found:
                char_not_found.add(char)

        return char_not_found
